const Gio = imports.gi.Gio;
const Lang = imports.lang;

const DBusProxy = Gio.DBusProxy.makeProxyWrapper(<interface name="org.freedesktop.DBus">
    <method name="GetNameOwner">
        <arg type="s" direction="in" />
        <arg type="s" direction="out" />
    </method>
    <method name="ListNames">
        <arg type="as" direction="out" />
    </method>
    <signal name="NameOwnerChanged">
        <arg type="s" direction="out" />
        <arg type="s" direction="out" />
        <arg type="s" direction="out" />
    </signal>
</interface>);

const PropertiesProxy = Gio.DBusProxy.makeProxyWrapper(<interface name="org.freedesktop.DBus.Properties">
    <signal name="PropertiesChanged">
        <arg type="s" direction="out" />
        <arg type="a{sv}" direction="out" />
        <arg type="as" direction="out" />
    </signal>
</interface>);

const MediaPlayer2Proxy = Gio.DBusProxy.makeProxyWrapper(<interface name="org.mpris.MediaPlayer2">
    <method name="Quit" />
    <method name="Raise" />
    <property name="CanQuit" type="b" access="read" />
    <property name="CanRaise" type="b" access="read" />
    <property name="CanSetFullscreen" type="b" access="read" />
    <property name="Fullscreen" type="b" access="readwrite" />
    <property name="HasTrackList" type="b" access="read" />
    <property name="Identity" type="s" access="read" />
    <property name="DesktopEntry" type="s" access="read" />
    <property name="SupportedMimeTypes" type="as" access="read" />
    <property name="SupportedUriSchemes" type="as" access="read" />
</interface>);
 
const MediaPlayer2PlayerProxy = Gio.DBusProxy.makeProxyWrapper(<interface name="org.mpris.MediaPlayer2.Player">
    <method name="PlayPause" />
    <method name="Pause" />
    <method name="Play" />
    <method name="Stop" />
    <method name="Next" />
    <method name="OpenUri">
        <arg type="s" direction="in" />
    </method>
    <method name="Previous" />
    <method name="Seek">
        <arg type="x" direction="in" />
    </method>
    <method name="SetPosition">
        <arg type="o" direction="in" />
        <arg type="x" direction="in" />
    </method>
    <property name="CanGoNext" type="b" access="read" />
    <property name="CanGoPrevious" type="b" access="read" />
    <property name="CanPause" type="b" access="read" />
    <property name="CanPlay" type="b" access="read" />
    <property name="CanSeek" type="b" access="read" />
    <property name="CanControl" type="b" access="read" />
    <property name="LoopStatus" type="s" access="readwrite" />
    <property name="MaximumRate" type="d" access="read" />
    <property name="Metadata" type="a{sv}" access="read" />
    <property name="MinimumRate" type="d" access="read" />
    <property name="PlaybackStatus" type="s" access="read" />
    <property name="Position" type="x" access="read" />
    <property name="Rate" type="d" access="readwrite" />
    <property name="Shuffle" type="b" access="readwrite" />
    <property name="Volume" type="d" access="readwrite" />
    <signal name="Seeked">
        <arg type="x" direction="out" />
    </signal>
</interface>);

const MediaPlayer2PlaylistsProxy = Gio.DBusProxy.makeProxyWrapper(<interface name="org.mpris.MediaPlayer2.Playlists">
    <method name="ActivatePlaylist">
        <arg type="o" direction="in" />
    </method>
    <method name="GetPlaylists">
        <arg type="u" direction="in" />
        <arg type="u" direction="in" />
        <arg type="s" direction="in" />
        <arg type="b" direction="in" />
        <arg type="a{oss}" direction="out" />
    </method>
    <property name="PlaylistCount" type="u" access="read" />
    <property name="Orderings" type="as" access="read" />
    <property name="ActivePlaylist" type="(b(oss))" access="read" />
    <signal name="PlaylistChanged">
        <arg type="oss" direction="out" />
    </signal>
</interface>);

const MediaPlayer2TrackListProxy = Gio.DBusProxy.makeProxyWrapper(<interface name="org.mpris.MediaPlayer2.TrackList">
    <method name="AddTrack">
        <arg type="s" direction="in" />
        <arg type="o" direction="in" />
        <arg type="b" direction="in" />
    </method>
    <method name="GetTracksMetadata">
        <arg type="ao" direction="in" />
        <arg type="aa{sv}" direction="out" />
    </method>
    <method name="GoTo">
        <arg type="o" direction="in" />
    </method>
    <method name="RemoveTrack">
        <arg type="o" direction="in" />
    </method>
    <property name="CanEditTracks" type="b" access="read" />
    <property name="Tracks" type="ao" access="read" />
    <signal name="TrackAdded">
        <arg type="a{sv}" direction="out" />
        <arg type="o" direction="out" />
    </signal>
    <signal name="TrackListReplaced">
        <arg type="ao" direction="out" />
        <arg type="o" direction="out" />
    </signal>
    <signal name="TrackMetadataChanged">
        <arg type="o" direction="out" />
        <arg type="a{sv}" direction="out" />
    </signal>
    <signal name="TrackRemoved">
        <arg type="o" direction="out" />
    </signal>
</interface>);

function DBus() {
    return new DBusProxy(Gio.DBus.session, 'org.freedesktop.DBus', '/org/freedesktop/DBus');
}

function Properties(owner) {
    return new PropertiesProxy(Gio.DBus.session, owner, '/org/mpris/MediaPlayer2');
}

function MediaPlayer2(owner) {
    return new MediaPlayer2Proxy(Gio.DBus.session, owner, '/org/mpris/MediaPlayer2');
}

function MediaPlayer2Player(owner) {
    return new MediaPlayer2PlayerProxy(Gio.DBus.session, owner, '/org/mpris/MediaPlayer2');
}

function MediaPlayer2Playlists(owner) {
    return new MediaPlayer2PlaylistsProxy(Gio.DBus.session, owner, '/org/mpris/MediaPlayer2');
}

function MediaPlayer2TrackList(owner) {
    return new MediaPlayer2TrackListProxy(Gio.DBus.session, owner, '/org/mpris/MediaPlayer2');
}
