/* Imports */
const _ = imports.gettext.domain('gnome-shell-extensions-mpris22mediaplayer').gettext;
const GObject = imports.gi.GObject;
const Gtk = imports.gi.Gtk;

/* Extension Imports */
const Me = imports.misc.extensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

/* Constants */
const mimes = ["audio/x-flac", "audio/mp4", "audio/mpeg", "audio/ogg", "audio/x-vorbis+ogg", "audio/x-ms-wma"];

/* Global Variables */
let settings;
let sound;
let timeAlwaysShown;

function init() {
}

function buildPrefsWidget() {
    settings = Convenience.getSettings();
    var frame = new Gtk.Box({orientation: Gtk.Orientation.VERTICAL});
    var table = new Gtk.Grid();
    var switch_menu = new Gtk.Switch({active: settings.get_boolean("sound-menu")});
    switch_menu.connect('notify::active', function(button) {
        settings.set_boolean("sound-menu", button.active);
    });
    var switch_time = new Gtk.Switch({active: settings.get_boolean("always-show-time")});
    switch_time.connect('notify::active', function(button) {
        settings.set_boolean("always-show-time", button.active);
    });
    var entry_mime = Gtk.ComboBoxText.new_with_entry();
    setEntries(entry_mime, mimes, "default-player-mime");
    entry_mime.connect("changed", function() {
        settings.set_string("default-player-mime", entry_mime.get_active_text());
    });
    table.attach(new Gtk.Label({label:_("Sound Menu"), xalign: 0, hexpand: true}), 0, 0, 1, 1);
    table.attach(switch_menu, 1, 0, 1, 1);
    table.attach(new Gtk.Label({label:_("Always Show Time"), xalign: 0, hexpand: true}), 0, 1, 1, 1);
    table.attach(switch_time, 1, 1, 1, 1);
    table.attach(new Gtk.Label({label:_("Default Media Player Mime Type"), xalign: 0, hexpand: true}), 0, 2, 1, 1);
    table.attach(entry_mime, 1, 2, 1, 1);
    frame.add(table);
    frame.show_all();
    settings.connect('changed', function (gsettings, key) {
        if (key=="sound-menu")
            switch_menu.set_active(settings.get_boolean(key));
        if (key=="always-show-time")
            switch_time.set_active(settings.get_boolean(key));
        if (key=="default-player-mime") {
            var mime = settings.get_string(key);
            if (entry_mime.get_active_text()!=mime) {
                setEntries(entry_mime, mimes, key);
                if (mimes.indexOf(mime)>=0)
                    entry_mime.set_active(mimes.indexOf(mime));
                else {
                    entry_mime.append_text(mime);
                    entry_mime.set_active(mimes.length);
                }
            }
        }
    });
    return frame;
}

function setEntries(entry, entries, key) {
    entry.remove_all();
    for (index in entries) {
        entry.append_text(entries[index]);
        if (settings.get_string(key)==entries[index])
            entry.set_active(index);
    }
}
